# python-jpl-horizon
#
# Written by: Matthew Mihok (@mattmattmatt)
#
# The MIT License (MIT)
# Copyright (c) 2013
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the
# following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import re
import datetime
from telnetlib import Telnet

from horizon.config import (
    HORIZON_HOST,
    HORIZON_PORT,
    HORIZON_PROMPT,
    HORIZON_QUERY_PROMPT,
    HORIZON_OPTION_PROMPT,
    HORIZON_MISC_PROMPT,
    HORIZON_CARTESIAN_PROMPT,
    DEBUG,
)

HORIZON_OBSERVE = 0
HORIZON_ELEMENTS = 1
HORIZON_VECTORS = 2

HORIZON_MAJOR_ALL = 0
HORIZON_MAJOR_PLANET = 1
HORIZON_MAJOR_MOON = 2
HORIZON_MAJOR_ARTIFICAL = 3


class Interface():
    telnet = Telnet()
    re_version = re.compile("version|v\s?(\d+\.\d+)")
    # re_major = re.compile("^\s+(-?\d+)\s\s(\w+)", flags=re.MULTILINE)
    re_major = re.compile("^\s{2}(.{7})\s{2}(.{34})\s(.{11})\s{2}(.*)$", flags=re.MULTILINE)
    re_minor = re.compile("^\s{4}(.{8})\s{2}(.{8})\s{2}(.{13})\s{2}(.{25})\s(.+)$", flags=re.MULTILINE)
    re_meta = re.compile("(\w+)\s+([\d\/\s]+)$", flags=re.MULTILINE)
    re_meta_dictA = re.compile("^\s{2}([A-Z].{0,21})=\s{1,2}(.{015})\s[A-Z]", flags=re.MULTILINE)
    re_meta_dictB = re.compile("\s([A-Z].{0,21})=\s{1,2}(.{0,16})\s?$", flags=re.MULTILINE)
    re_meta_fail = re.compile("No such object record found\: \d+")
    re_cartesian = re.compile("\$\$SOE(.*)\$\$EOE", flags=re.DOTALL)

    def __open(self):
        self.telnet.open(HORIZON_HOST, HORIZON_PORT)
        self.telnet.read_until(HORIZON_PROMPT)
        # self.telnet.read_until(HORIZON_PROMPT)

    def __close(self, send_quit=False):
        if send_quit:
            self.telnet.write("\rquit\n")

        self.telnet.close()

    def __parse_major(self, data, major_type=HORIZON_MAJOR_ALL):
        matches = self.re_major.findall(data)
        # import pdb; pdb.set_trace()
        if len(matches) is 0:
            return []

        if major_type == HORIZON_MAJOR_ALL:
            matches = list(dict({
                'id': match[0].strip(),
                'name': match[1].strip(),
                'designation': match[2].strip(),
                'alias': match[3].strip(),
            }) for match in matches)
        else:
            buff = list()

            for match in matches:
                id = 0
                # skip match if id is not an integer
                try:
                    id = int(match[0])
                except ValueError:
                    continue

                # import pdb; pdb.set_trace()
                if major_type is HORIZON_MAJOR_ARTIFICAL and (id < 0 or id > 1000):
                    # print "SATILLITE {0}".format(match[0])
                    buff.append(dict({
                        'id': match[0].strip(),
                        'name': match[1].strip(),
                        'designation': match[2].strip(),
                        'alias': match[3].strip(),
                    }))

                elif (major_type is HORIZON_MAJOR_PLANET and
                        id % 100 == 99 and
                        id > 10):
                    # print "PLANET {0}".format(match[0])
                    buff.append(dict({
                        'id': match[0].strip(),
                        'name': match[1].strip(),
                        'designation': match[2].strip(),
                        'alias': match[3].strip(),
                    }))

                elif (major_type is HORIZON_MAJOR_MOON and
                        id % 10 != 9 and
                        id % 10 != 0 and
                        id > 10 and
                        id < 1000):
                    # print "MOON {0} ({0} % 10 == {1})".format(match[0], id % 10)
                    buff.append(dict({
                        'id': match[0].strip(),
                        'name': match[1].strip(),
                        'designation': match[2].strip(),
                        'alias': match[3].strip(),
                    }))
                else:
                    continue
            matches = buff

        return matches

    def __parse_minor(self, data):
        matches = self.re_minor.findall(data)

        if len(matches) is 0:
            return []

        # flip the results
        # matches = list((m[2].lower(), m[0], m[1], m[3]) for m in matches)

        matches = list(dict({
            'id': match[0].strip(),
            'epoch': match[1].strip(),
            'designation': match[2].strip(),
            'name': match[3].strip(),
            'radius': match[4].strip(),
        }) for match in matches)

        # print matches

        return matches

    def __parse_meta(self, data):

        if self.re_meta_fail.match(data):
            return {'error': 'No such object record found'}

        matches = []
        matches += self.re_meta_dictA.findall(data)
        matches += self.re_meta_dictB.findall(data)
        meta = self.re_meta.search(data).groups()
        matches = list(tuple(v.strip() for v in m) for m in matches)

        # import pdb; pdb.set_trace()

        name = meta[0]
        id = meta[1].strip().split(' / ')[0]

        # include name and id along with detailed data
        matches += [('ID', id)]
        matches += [('Name', name)]

        return dict(matches)

    def __parse_cartesian(self, data):
        matches = self.re_cartesian.search(data).groups()
        matches = [m.strip().split("\r\n") for m in matches][0]
        result = []

        for i in range(len(matches)):
            vector = dict()
            key = ""
            if (i / 4) % 4 == 0:
                key = "time"
            elif (i / 4) % 4 == 3:
                key = "ltrgrr"
            elif (i / 4) % 4 == 2:
                key = "vxvyvz"
            elif (i / 4) % 4 == 1:
                key = "xyz"
            else:
                break
                raise Exception()
            vector[key] = matches[i]
            result.append(vector)
        return result

    def __parse_version(self, data):
        matches = self.re_version.search(data)

        if matches is None:
            return 0
        # fail if cant find
        return matches.groups()[0]

    def get_major(self, type=HORIZON_MAJOR_ALL):
        """
        =================
        get_major ( self, type ) : returns [{},]
        =================

        type: horizon.interface.HORIZON_MAJOR_ALL (default)
              horizon.interface.HORIZON_MAJOR_ARTIFICAL
              horizon.interface.HORIZON_MAJOR_MOON
              horizon.interface.HORIZON_MAJOR_PLANET

        Returns a list of major bodies and corresponding IDs
        """
        self.__open()
        self.telnet.write("MB\n")
        # import pdb;pdb.set_trace()

        result = self.telnet.read_until(HORIZON_QUERY_PROMPT)

        self.__close(send_quit=True)

        return self.__parse_major(result, type)

    def get_minor(self, page=0, page_size=-1):
        """
        =================
        get_minor ( self, page ) : returns [{},]
        =================

        page: (default: 0) integer to specify page in pagination

        page_size (default: -1): integer to specify how many
            results per page

        Returns a list of minor bodies and corresponding IDs
        """
        self.__open()
        self.telnet.write("RAD > 0\n")

        self.telnet.read_until(HORIZON_MISC_PROMPT)
        # print buff

        self.telnet.write("\r\n")
        result = self.telnet.read_until(HORIZON_QUERY_PROMPT)
        # print result

        self.__close(send_quit=True)

        return self.__parse_minor(result)

    def get_cartesian(self, id, start, end, ref="500@10", type=HORIZON_VECTORS, frequency="1h"):

        # make sure start / end are valid or throw error

        self.__open()
        self.telnet.write("{0}\n".format(id))

        self.telnet.read_until(HORIZON_QUERY_PROMPT)
        # print buff

        self.telnet.write("E\n")
        self.telnet.read_until(HORIZON_MISC_PROMPT)
        # print buff

        # select cartesian data type
        if type is HORIZON_OBSERVE:
            self.telnet.write("o\n")
        elif type is HORIZON_ELEMENTS:
            self.telnet.write("e\n")
        else:
            self.telnet.write("v\n")

        self.telnet.read_until(HORIZON_MISC_PROMPT)

        # select reference point ID, coord, geo
        self.telnet.write("{0}\n".format(ref))
        self.telnet.read_until(HORIZON_OPTION_PROMPT)
        self.telnet.write("y\n")

        # select reference plane: body, eclip, frame
        self.telnet.read_until(HORIZON_MISC_PROMPT)
        self.telnet.write("body\n")

        self.telnet.read_until(HORIZON_MISC_PROMPT)
        self.telnet.write("{0}\n".format(start))
        self.telnet.read_until(HORIZON_MISC_PROMPT)
        self.telnet.write("{0}\n".format(end))

        # frequency: 1d, 1h, 1m
        self.telnet.read_until(HORIZON_MISC_PROMPT)
        self.telnet.write("{0}\n".format(frequency))

        # accept output
        self.telnet.read_until(HORIZON_MISC_PROMPT)
        self.telnet.write("\r\n")

        result = self.telnet.read_until(HORIZON_CARTESIAN_PROMPT)

        # Cartesian queries have a weird exit
        self.telnet.write("N\n")
        self.__close(send_quit=True)
        return self.__parse_cartesian(result)

    def query(self, query):
        self.__open()
        self.telnet.write("{0}\n".format(query))

        result = self.telnet.read_until(HORIZON_QUERY_PROMPT)

        self.__close(send_quit=True)

        if "Number of matches" in result:
            return self.__parse_list(result)

        # if query returns single result, use meta parser
        return self.__parse_meta(result)

    def get(self, id):
        self.__open()
        self.telnet.write("{0}\n".format(id))

        result = self.telnet.read_until(HORIZON_QUERY_PROMPT)

        self.__close(send_quit=True)

        return self.__parse_meta(result)

    def get_version(self):
        self.__open()
        self.telnet.write("quit\n")

        result = self.telnet.read_until(HORIZON_QUERY_PROMPT)

        self.__close()

        return self.__parse_version(result)



HORIZON_HOST = "horizons.jpl.nasa.gov"
HORIZON_PORT = 6775
HORIZON_PROMPT = "Horizons>"
HORIZON_DEFAULT_PROMPT = ":"
HORIZON_QUERY_PROMPT = "<cr>:"
HORIZON_OPTION_PROMPT = " --> "
HORIZON_CARTESIAN_PROMPT = "? : "
HORIZON_MISC_PROMPT = "] : "
DEBUG = False



from distutils.core import setup

setup(
    name='HorizonJPL',
    version='0.1.7',
    author='Matthew Mihok',
    author_email='matthew.mihok@niewma.com',
    packages=['horizon', 'horizon.test'],
    scripts=[],
    url='http://pypi.python.org/pypi/HorizonJPL/',
    license='LICENSE.txt',
    description='Access to NASA\'s JPL HORIZON Ephemeris System',
    
long_description='{0}\r\n\r\nChanges\r\n-----------------\r\n{1}'.format(open('README.txt').read(),open('CHANGES.txt').read()),
    install_requires=[
        'pytest-timeout >= 0.3'
    ]
)



# python-jpl-horizon
#
# Written by: Matthew Mihok (@mattmattmatt)
#
# The MIT License (MIT)
# Copyright (c) 2013
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the
# following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import unittest
import pytest

from horizon.interface import Interface


class TestHorizonInterface(unittest.TestCase):
    # print "Running Tests..."

    def setUp(self):
        self.horizon = Interface()

    @pytest.mark.timeout(10)
    def test_connection(self):
        version = self.horizon.get_version()

        self.assertTrue(version)

    @pytest.mark.timeout(300)
    def test_minor_bodies(self):
        data = self.horizon.get_minor()

        self.assertTrue(len(data))

    @pytest.mark.timeout(10)
    def test_major_bodies(self):
        data = self.horizon.get_major()

        self.assertTrue(len(data))

    @pytest.mark.timeout(10)
    def test_get(self):
        data = self.horizon.get(10)

        self.assertTrue(data)



# python-jpl-horizon
#
# Written by: Siddarth Kalra (@siddarthkalra)
#
# The MIT License (MIT)
# Copyright (c) 2013
#
# Permission is hereby granted, free of charge, to any person obtaining a
# copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to permit
# persons to whom the Software is furnished to do so, subject to the
# following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
# OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
# THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.

import sys
import json
from os import curdir, sep

import SimpleHTTPServer
import SocketServer
from BaseHTTPServer import BaseHTTPRequestHandler

from horizon.interface import Interface
import urllib
import re

class JplRequestHandler(BaseHTTPRequestHandler):
        
    #TODO -- more convenience API
    #by enum and name --> planet, asteroid, spacecraft, moon --> .get(name, enum)
    #by name --> .get(name)
        
    def do_GET(self):
    
        print "PATH: " + self.path
        
        #determine the request type being received
        req_info = self.__get_request_info(self.path)
        if req_info["type"] is None:
            self.__send_http_response_400("Invalid request type.")
            return
        
        response = {
            "version": None
        }        
    
        if req_info["type"] == "id_req":
            #initialize horizon
            horizon_data = Interface()
            response["version"] = horizon_data.version()
                        
            if req_info["params"] is None or len(req_info["params"]) == 0:
                self.__send_http_response_400("Invalid request, no body id specified.")
                return
            elif not req_info["params"].isdigit():
                self.__send_http_response_400("Invalid request, body id must be an integer.")
                return
            
            response["body_id"] = horizon_data.get(req_info["params"])
            
            #set success headers
            self.__send_http_response_200("application/json")
                
            #do the JSON magic
            json.dump({"horizons-api": response}, self.wfile)
        elif req_info["type"] == "name_req":
            #initialize horizon
            horizon_data = Horizon()
            response["version"] = horizon_data.version()
                        
            if req_info["params"] is None or len(req_info["params"]) == 0:
                self.__send_http_response_400("Invalid request, no body name specified.")
                return
            
            response["body_name"] = horizon_data.get(req_info["params"])
            
            #set success headers
            self.__send_http_response_200("application/json")
                
            #do the JSON magic
            json.dump({"horizons-api": response}, self.wfile)
        elif req_info["type"] == "list_req":
            #initialize horizon
            horizon_data = Horizon()
            response["version"] = horizon_data.version()
                        
            if req_info["params"] is None or len(req_info["params"]) == 0:
                self.__send_http_response_400("Invalid request, no list type specified.")
                return
            elif req_info["params"] != "mb" and req_info["params"] != "sb":
                self.__send_http_response_400("Invalid request, only mb or sb lists are supported.")
                return
            
            if req_info["params"] == "mb":
                response["mb"] = horizon_data.major()
            elif req_info["params"] == "sb":
                response["mb"] = horizon_data.minor()
            
            #set success headers
            self.__send_http_response_200("application/json")
                
            #do the JSON magic
            json.dump({"horizons-api": response}, self.wfile)      
        elif req_info["type"] == "complex_req":
            json_obj = self.__get_complex_query_json(self.path)
            
            if json_obj is None:
                self.__send_http_response_400("Invalid json provided.")
                return
            
            #validate the json request 
            json_validation = self.__is_json_obj_valid(json_obj)
            print json_validation
            
            if not json_validation["success"]:
                self.__send_http_response_400(json_validation["message"])
            else:
                json_obj = json_obj["horizons-api"]
                
                #initialize horizon
                horizon_data = Horizon()
                response["version"] = horizon_data.version()
                
                #decision tree for what horizon method to run
                if json_obj["query_type"] == "list":
                    if json_obj["filters"]["body_type"] == "mb":
                        response["mb"] = horizon_data.major() 
                    elif json_obj["filters"]["body_type"] == "sb":
                        response["sb"] = horizon_data.minor()
                elif json_obj["query_type"] == "body_id":
                    response["body_id"] = horizon_data.get(json_obj["filters"]["value"])
                elif json_obj["query_type"] == "body_name": 
                    response["body_name"] = horizon_data.get(json_obj["filters"]["value"])
            
                #set success headers
                self.__send_http_response_200("application/json")
                
                #do the JSON magic
                json.dump({"horizons-api": response}, self.wfile)
        elif req_info["type"] == "demo_req":
            f = open("/data/web/nasa.api.niewma.com/python-jpl-horizon/demo/index.html")
            #note that this potentially makes every file on your computer readable by the internet

            self.__send_http_response_200("text/html")
            self.wfile.write(f.read())
            f.close()
        else:
            f = open("/data/web/nasa.api.niewma.com/python-jpl-horizon/demo/" + self.path)
            self.send_response(200)
            self.send_header('Content-type',    'text/javascript')
            self.end_headers()

        # elif req_info["type"] == "file_req":
        #     f = open(curdir + sep + "demo" + self.path)
        #     self.__send_http_response_200("text/html")

            self.wfile.write(f.read())
            f.close()
            
        return
        
    #PRIVATE methods

    "determine the request type being received"
    def __get_request_info(self, path):
        demo_request = re.compile("^\/$")
        file_request = re.compile("^\/(.+)\.{1}[a-z]+$")
        id_request = re.compile("^\/api\?body_id=(.*)$")
        name_request = re.compile("^\/api\?body_name=(.*)$")
        list_request = re.compile("^\/api\?list=(.*)$")
        query_request = re.compile("^\/api\?query=(.*)$")
        
        request_info = {
            "params": None,
            "type": None 
        }
        
        matches = file_request.search(path)
        
        if matches is not None:
            request_info["type"] = "file_req"
            
            return request_info
        
        matches = demo_request.search(path)
        
        if matches is not None:
            request_info["type"] = "demo_req"
            
            return request_info
                    
        matches = query_request.search(path)
        
        if matches is not None:
            query_json = matches.groups()[0]
            request_info["params"] = query_json
            request_info["type"] = "complex_req"
            
            return request_info
        
        matches = id_request.search(path)
        
        if matches is not None:
            id = matches.groups()[0]
            request_info["params"] = id
            request_info["type"] = "id_req"
            
            return request_info
        
        matches = name_request.search(path)
        
        if matches is not None:
            name = matches.groups()[0]
            request_info["params"] = name
            request_info["type"] = "name_req"
            
            return request_info
        
        matches = list_request.search(path)
        
        if matches is not None:
            list_value = matches.groups()[0]
            request_info["params"] = list_value
            request_info["type"] = "list_req"
            
            return request_info
        
        return request_info

    def __send_http_response_200(self, content_type):
        self.send_response(200)
        self.send_header("Content-type", content_type)
        self.end_headers()

    "send back a bad request header with a message"
    def __send_http_response_400(self, message):
        self.send_response(400)
        self.send_header("Content-type", "text/html")
        self.end_headers()
        self.wfile.write(message)        

    "read the url path and load the JSON for a complex"
    def __get_complex_query_json(self, path):
        json_str = urllib.unquote_plus(path[11:])
        print json_str

        try:
            return json.loads(json_str)
        except:
            return None

    "validate the json request object"
    def __is_json_obj_valid(self, json_obj):
        required_keys = ("version", "response_type", "query_type", "filters")
        query_types = ("id", "name", "list") #to be used
        list_body_types = ("mb", "sb") #to be used
        response_types = ("json") #to be used

        result = {"success": False, "message": "Failed to validate request."}

        if not "horizons-api" in json_obj:
            result["message"] = "Failed. First level key 'horizons-api' missing."
            return result

        for r_key in required_keys:
            if not r_key in json_obj["horizons-api"]:
                result["message"] = "Failed. Second level key '%s' missing." % r_key
                return result

         #TODO -- more in-depth JSON query validation
         #if json_obj["horizons-api"]["query_type"] == "list":
         #elif json_obj["horizons-api"]["query_type"] == "id":
         #elif json_obj["horizons-api"]["query_type"] == "name":

        result["success"] = True
        result["message"] = "Success"
        return result

#HTTP server related code
if sys.argv[1:]:
    port = int(sys.argv[1])
else:
    port = 8000
server_address = ('127.0.0.1', port)

httpd = SocketServer.TCPServer(server_address, JplRequestHandler)

print "serving at port", port
httpd.serve_forever()
